-- Time-stamp: <Wed Jan 18 2006 15:23:08 hwloidl>
--
-- Interpreter for the Intermediate Language of the MiniJava compiler
---------------------------------------------------------------------------

module Interp where

import AbsSyn
import PPIntm
--import Parse
import Data.Map as Map
import Control.Monad
import Text.Regex

import Debug.Trace(trace)

type RegMap = Map Register Val
type Frame = Integer
type Offset = Integer
type Memory = Map Address Val
type Address = Integer
type Trace = String
data Val = I !Integer |
           A !Address |
           L !Label
     deriving (Read,Show,Eq)
type Msgs = [String]

data TraceOpts = CallTrace | ParamTrace | JumpTrace | MoveTrace | AllocTrace | RegMaps | MemMaps | LabMaps | StdCallConv | Warnings | Debug | Verbose | None
 deriving (Read,Show,Eq)
type Settings = ([TraceOpts])

type LabTab = Map Label [Stm]
type FunTab = Map Label [Stm]
type FunParams = Map Label [Access]
type SpecialRegs = (String,String,String,String,Integer,Integer)

type State = (RegMap,Memory,LabTab,FunTab,FunParams,Msgs,Settings,Cost,SpecialRegs)

-- evaluation monad
data Intp a = T (State -> (State, a))

runIntp :: State -> Intp a -> a
runIntp s (T f) = snd (f s)

instance Monad Intp where
 return x = T (\ s -> (s, x))
 (T t) >>= f =  T (\ s -> let { (s1,x1) = t s ; T f' = f x1 } in f' s1)

instance Num Val where
 (A x) + (A y) = A (x+y)
 (A x) + (I y) = A (x+y)
 (I x) + (A y) = A (x+y)
 (I x) + (I y) = I (x+y)
 (A x) * (A y) = A (x*y)
 (A x) * (I y) = A (x*y)
 (I x) * (A y) = A (x*y)
 (I x) * (I y) = I (x*y)
 abs (I x) = I (abs x)
 abs (A _) = error "Cannot apply abs to an address"
 signum (I x) = I (signum x)
 signum (A _) = error "Cannot apply signum to an address"
 fromInteger n = I n

instance Ord Val where
 (A x) <= (A y) = x<=y
 (I x) <= (I y) = x<=y
 _ <= _ = error "Cannot compare address with integer"
 
emptyEnv :: RegMap
emptyEnv = empty

emptyTrace :: Msgs
emptyTrace = []

defaultSettings :: Settings
defaultSettings = ([CallTrace,JumpTrace,MoveTrace,AllocTrace,ParamTrace,LabMaps])

defaultRegs = ("t32","t9998","t9999","t1",0,0)

emptyState :: [TraceOpts] -> SpecialRegs -> State
emptyState topts regs = (Map.empty, Map.empty, 
              	    Map.empty, Map.empty, Map.empty,
              	    emptyTrace, topts, 0, regs)

combineStates (rm, mm, lT, fT, pT, ts, ss, c, r) (rm', mm', lT', fT', pT', ts', ss', c', _) =
  (Map.union rm rm', Map.union mm mm', Map.union lT lT', Map.union fT fT', Map.union pT pT', ts++ts', ss++ss', c+c', r)

-- type State = (RegMap,Memory,LabTab,FunTab,FunParams,Msgs,Settings)
initState :: [TraceOpts] -> SpecialRegs -> [AbsSyn.Fct] -> State
initState topts regs = (foldl combineStates (emptyState topts regs)) .
            (Prelude.map ( \ (Fct name ps ss) -> (Map.empty, Map.empty, -- empty regs and heap
	                      runLabTabFiller (fillTab ss),
                              Map.union Map.empty (Map.singleton name ss), -- body
                              Map.union Map.empty (Map.singleton name (Prelude.map param2access ps)), -- params
                              emptyTrace, [], 0, ("","","","")) ))

-- return value of functions
retReg :: String 
retReg = "t2"

retRegA :: Access
retRegA = InReg retReg

-- names of callee save registers (UNUSED)
calleeSaves :: [String]
calleeSaves = ["t40","t41","t42","t43"]

thisReg :: Intp String 
thisReg =  T $ \ st@(e,m,lT,fT,fP,t,s,c,(_,_,_,tp,_,_)) -> (st,tp)

-- thisRegA :: Access
-- thisRegA = InReg thisReg

fpReg :: Intp String 
fpReg = T $ \ st@(e,m,lT,fT,fP,t,s,c,(fp,_,_,_,_,_)) -> (st,fp)

spReg :: Intp String 
spReg = T $ \ st@(e,m,lT,fT,fP,t,s,c,(_,sp,_,_,_,_)) -> (st,sp)

hpReg :: Intp String 
hpReg = T $ \ st@(e,m,lT,fT,fP,t,s,c,(_,_,hp,_,_,_)) -> (st,hp)

spSize :: Intp Integer
spSize = T $ \ st@(e,m,lT,fT,fP,t,s,c,(_,_,_,_,sz,_)) -> (st,sz)

hpSize :: Intp Integer
hpSize = T $ \ st@(e,m,lT,fT,fP,t,s,c,(_,_,_,_,_,sz)) -> (st,sz)

-- dummy vals; mainly to catch errors of uninit vars
dummyInt = (I 9996)
dummyAddr = (A 0)

dummyRegVal = (I 8881)
dummyMemVal = (I 8882)

spStart = (A 8000)
hpStart = (A 9000)
spEnd = (A 1024)
hpEnd = (A 64*65535)
defHpSz = 16*65536

globalFrameSizeHACK = 20

---

showRegMap :: Intp ()
showRegMap = getR >>= \ rm -> trace ("Register Mapping: " ++ (showTree rm)) (return ())
--showRegMap = getR >>= \ rm -> msgs ["Register Mapping: ", (showTree rm)]

showMemMap :: Intp ()
showMemMap = getM >>= \ rm -> trace ("Memory Mapping: " ++ (showTree rm)) (return ())
--showMemMap = getM >>= \ rm -> msgs ["Memory Mapping: ", (showTree rm)]

showLabTab :: Intp ()
showLabTab = labTab >>= \ lT -> trace ("labTab: " ++ (showTreeWith (\ k a -> k ++ ":=" ++ render (pp a)) True False lT)) (return ())
--showLabTab = labTab >>= \ lT -> msgs ["labTab: ", (showTreeWith (\ k a -> k ++ ":=" ++ render (pp a)) True False lT)]

showFunTab :: Intp ()
showFunTab = funTab >>= \ fT -> trace ("funTab: " ++ ((showTreeWith (\ k a -> k ++ ":=" ++ render (pp a)) False False fT))) (return ())
--showFunTab = funTab >>= \ fT -> msgs ["funTab: ", ((showTreeWith (\ k a -> k ++ ":=" ++ render (pp a)) False False fT))]

showParamTab :: Intp ()
showParamTab = paramTab >>= \ fT -> trace ("paramTab: " ++ ((showTreeWith (\ k a -> k ++ ":=" ++ (show a)) False False fT))) (return ())

execCall :: [TraceOpts] -> SpecialRegs -> [AbsSyn.Fct] -> AbsSyn.Fct -> [Integer] -> ([String],Cost,Int,Val)
execCall topts regs  prg (Fct name ps ss) es = 
  let
       state = initState topts regs prg
       code = do
                --when (Interp.Verbose `elem` topts) $
                --  trace ("++ initState with special registers " ++ (show regs) ++ " and trace opts " ++ (show topts)) (return ())
                when (Interp.Debug `elem` topts) $
                 (showLabTab >> showFunTab >> showParamTab)
                rm <- initRegs regs
                setR rm
                rm <- getR 
                when (Interp.Verbose `elem` topts) $
                  trace ("Regs after init: " ++ (showTree rm)) (return ())
                -- trace ("Executing: " ++ (show (CALL (NAME name) ((TEMP thisReg):(Prelude.map (\ n -> CONST n) es))))) (return ())
                tp <- thisReg
                res <- eval (CALL (NAME name) ((TEMP tp):(Prelude.map (\ n -> CONST n) es)))
                m <- getM
                t <- getT
                c <- getClock
                return (reverse t,
                        c,length (Map.keys m),
                        res)

     in
       runIntp state code
	                             

dummyVal :: Val
dummyVal = I 99

msg :: String -> Intp ()
msg str = T $ \ (e,m,lT,fT,fP,t,s,c,p) -> ((e,m,lT,fT,fP,str:t,s,c,p), ())

msgs :: [String] -> Intp ()
msgs xs = sequence_ (Prelude.map msg xs)

getR = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,e)
getM = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,m)
getT = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,t)
getS = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,s)
getSpecials = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,p)
getClock = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,c)
labTab = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,lT)
funTab = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,fT)
paramTab = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,fP)
funParams = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st,fP)

setR x = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> ((x,m,lT,fT,fP,t,s,c,p),())
setM x = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> ((e,x,lT,fT,fP,t,s,c,p),())
setT x = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> ((e,m,lT,fT,fP,x,s,c,p),())

getReg :: Register -> Intp Val
getReg r = getR >>= \ e -> let x = Map.findWithDefault dummyRegVal r e in return x

getRegSafe :: Register -> Intp Val
getRegSafe a = getReg a >>= \ x -> if x==dummyMemVal then error ("PANIC: lookup to uninitialised register " ++ (show a)) else return x

setReg :: Register -> Val -> Intp ()
setReg r x = getR >>= \ rm -> setR (Map.union (Map.singleton r x) rm)

getMem :: Address -> Intp Val
getMem a = getM >>= \ m -> let x = Map.findWithDefault dummyMemVal a m in  return x
--getMem a = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> (st, Map.findWithDefault 81 a m)

getMemSafe :: Address -> Intp Val
getMemSafe a = getMem a >>= \ x -> if x==dummyMemVal then error ("PANIC: lookup to uninitialised mem cell " ++ (show a)) else return x

getLabCode :: Label -> Intp [Stm]
getLabCode l = labTab >>= \ lT -> let x = Map.findWithDefault [] l lT in  return x

getFunCode :: String -> Intp [Stm]
getFunCode f = funTab >>= \ fT -> let x = Map.findWithDefault [] f fT in  return x

getFunParams :: String -> Intp [Access]
getFunParams f = funParams >>= \ fP -> let x = Map.findWithDefault [] f fP in  return x

getTraceOpts :: Intp [TraceOpts]
getTraceOpts = getS >>= \ (topts) -> return topts

tick :: Cost -> Intp ()
tick n = T $ \ st@(e,m,lT,fT,fP,t,s,c,p) -> ((e,m,lT,fT,fP,t,s,c+n,p),())

rtError str = error ("** PANIC: "++str)
warn str x = trace ("++ WARNING: "++str) x

type BState = (String, [Stm], Map String [Stm], [(String,[Stm])])
newtype LabTabFiller a = B (BState -> (BState, a))

addBB :: Stm -> LabTabFiller ()
addBB bb = B $ \ (currLab, currBB, m, ss) -> ((currLab, currBB++[bb], m, ss),())

storeCurrBB :: LabTabFiller ()
storeCurrBB = B $ \ (currLab, currBB, m, ss) ->
                       if (currLab/="")
                         then {- trace ("$$ nested label " ++ currLab ++ " added to labTab; code:" ++ (show currBB)) -} ((currLab,currBB,Map.union (Map.singleton currLab currBB) m, ss),()) 
                         else if (Prelude.null currBB)
                                then ((currLab,currBB,m,ss), ())
                                else {- trace ("PANIC: storeCurrBB nuking currBB: " ++ (show currBB)) -} ((currLab,currBB,m,ss), ())

-- like storeCurrBB, but if BB doesn't end with a JUMP or CJUMP, add a JUMP
safeStoreCurrBB :: Label -> LabTabFiller ()
safeStoreCurrBB l = getCurrBB >>= \ currBB ->
                    if (Prelude.null currBB) 
                      then addBB (JUMP (NAME l) [l]) >> storeCurrBB
                      else case (Prelude.last currBB) of
                            (JUMP _ _) -> storeCurrBB
                            (CJUMP _ _ _ _ _) -> storeCurrBB
		            _ -> addBB (JUMP (NAME l) [l]) >> storeCurrBB

getCurrBB = B $ \ (currLab, currBB, m, ss) -> ((currLab, currBB, m, ss), currBB)

initCtxt :: String -> LabTabFiller ()
initCtxt l = B $ \ (currLab, currBB, m, ss) -> ((l, [], m, ss), ())


pushCtxt :: LabTabFiller ()
pushCtxt = B $ \ (currLab, currBB, m, ss) -> (("", [], m, (currLab, currBB):ss), ())

popCtxt :: LabTabFiller ()
popCtxt = storeCurrBB >> (B $ \ (_, _, m, (l,bb):ss) -> ((l,bb,m,ss),()))

{-
                 if (currLab/="")
                   then trace ("$$ nested label " ++ currLab ++ " added to labTab; code:" ++ (show currBB)) ((l,bb,Map.union (Map.singleton currLab currBB) m, ss),()) 
                   else ((l,bb,m,ss), ())
-}

runLabTabFiller :: LabTabFiller a -> Map String [Stm]
runLabTabFiller (B f) = let ((_,_,m,_),_) = f ("", [], Map.empty, []) in m

instance Monad LabTabFiller where
 return x = B (\ s -> (s, x))
 (B t) >>= f =  B (\ s -> let { (s1,x1) = t s ; B f' = f x1 } in f' s1)

type  Cost = Integer -- in machine cycles


class Evaluator x where
 fillTab :: x -> LabTabFiller ()
 cost :: x -> Cost
 eval :: x -> Intp Val

 {-
 prints :: [x] -> IO ()
 prints [] = return ()
 prints (e1:es) = runIntp emptyState (eval e1 >>= \ (_,t1,_) -> eval es >>= \ (_,t2,_) -> return (t1++"\n"++t2++"\n")) >>= \ str -> putStrLn str
 -}

instance Evaluator Stm where
 eval c@(MOVE dest src) = do 
                           -- trace ("Entering " ++ (show c) ++"\n") return ()
                           v <- eval src
                           access <- lval dest
                           store access v
                           topts <- getTraceOpts 
                           let traceme = MoveTrace `elem` topts
                           let warnme = Warnings `elem` topts
                           let rmaps = RegMaps `elem` topts
                           let mmaps = MemMaps `elem` topts
                           when traceme $
                              trace (".. MOVE " ++ (show access) ++ " <- " ++ (show v))  return ()  
                           when rmaps $
                             case (access) of {
                               (InReg r) -> getR >>= \ regs -> trace (".. Registers after " ++ (render (pp c)) ++ "\n" ++ showTree regs ++"\n") (return ()) ;
                                (InMem _) -> return () }
                           when mmaps $
                             case (access) of {
                               (InReg r) -> return () ;
                               (InMem _) -> showAllMem >>= \ str -> trace ("All Mem Dump:\n" ++ str) (return ()) }
                           when warnme $
                            case v of 
                             x@(A a) -> if (x==dummyAddr) then warn ("MOVE: reading uninit Addr " ++ (show x) ++ " from mem " ++ (show access) ++ " in expr " ++ (show src) ) (return ()) else return ()
                             x@(I i) -> if (x==dummyInt) then warn ("MOVE: reading uninit Int " ++ (show x) ++ " from mem " ++ (show access) ++ " in expr " ++ (show src) ) (return ()) else return ()
                           tick (cost c)
                           return dummyVal

 eval c@(EXP e) = tick (cost c) >> eval e 


 eval c@(JUMP dest poss) = do  -- ToDo: test poss
                          (L lab) <- eval dest
                          stm <- getLabCode lab
                          topts <- getTraceOpts 
                          let traceme = JumpTrace `elem` topts
                          let maps = LabMaps `elem` topts
                          if traceme 
                            then trace ("-> JUMP to label " ++ lab ++ (if maps then " continues with code\n" ++ (show stm) else "")) return ()
                            else return ()
                          tick (cost c)
                          eval stm

 eval c@(CJUMP rel left right trueLab falseLab) =
   do
     x <- eval left
     y <- eval right
     let z = case rel of
              AbsSyn.EQ -> x==y
              AbsSyn.NE -> x/=y
              AbsSyn.LT -> x<y
              AbsSyn.GT -> x>y
              AbsSyn.LE -> x<=y
              AbsSyn.GE -> x>=y
              AbsSyn.ULT -> x<y
              AbsSyn.ULE -> x<=y
              AbsSyn.UGT -> x>y
              AbsSyn.UGE -> x>=y
     
     topts <- getTraceOpts 
     let traceme = JumpTrace `elem` topts
     let maps = LabMaps `elem` topts
     if traceme 
       then trace ("__ CJUMP results in " ++ (show x ) ++ " " ++ (show rel) ++ " " ++ (show y) ++ " which is " ++ (show z)++"\tjumping to label " ++ (if z then trueLab else falseLab)) return ()
       else  return ()
     tick (cost c)
     if z
       then getLabCode trueLab >>= \ c -> tick 1 {- jump only in true case -} >> eval c
       else getLabCode falseLab >>= \ c -> eval c

 -- NB: control-flow handling only works for SEQ-lists, not for arb trees!
 eval c@(SEQ first@(JUMP _ _) second) = do 
                                         tick (cost first)
                             	         eval first
 eval (SEQ first@(CJUMP _ _ _ _ _) second) = do 
                                              tick (cost first)
                             	              eval first
 eval c@(SEQ first second) = do 
                               tick (cost c)
                               eval first
                               eval second

 eval (LABEL l) =  return (L l)

 -- NB: cover expressions, too, since ESEQ might have labels in the first component
 fillTab this@(MOVE dest src) = addBB this >> pushCtxt >> fillTab dest >> popCtxt >>  pushCtxt >> fillTab src >> popCtxt
 fillTab this@(EXP e) = addBB this >>  pushCtxt >> fillTab e >> popCtxt
 fillTab this@(JUMP dest poss) = addBB this >>  pushCtxt >> fillTab dest >> popCtxt
 fillTab this@(CJUMP rel left right trueLab falseLab) = addBB this >>  pushCtxt >> fillTab left >> popCtxt >>  pushCtxt >> fillTab right >> popCtxt
 fillTab this@(SEQ first second) = fillTab first >> fillTab second
 fillTab this@(LABEL l) = safeStoreCurrBB l >> initCtxt l 

 -- Cost metric
 cost (MOVE dest src) = 1
 cost (EXP e) = cost e
 cost (JUMP dest poss) = 1
 cost (CJUMP rel left right trueLab falseLab) = cost left + cost right + 1{- rel -} {- + 1 if result is true; assumes fall-through for false -}
 cost (SEQ first second) = cost first + cost second{- unless it's a jump!! -}
 cost (LABEL l) = 0

intOp :: AbsSyn.BinOp -> Integer -> Integer -> Integer
intOp AbsSyn.PLUS x y  = x+y
intOp AbsSyn.MINUS x y  = x-y
intOp AbsSyn.MUL x y  = x*y
intOp AbsSyn.DIV x y  = x `div` y
intOp AbsSyn.AND x y  = boolToVal ((valToBool x)&&(valToBool y))
intOp AbsSyn.OR x y  = boolToVal ((valToBool x)||(valToBool y))
intOp AbsSyn.LSHIFT x y  = x*(mypow 2 y)
intOp AbsSyn.RSHIFT x y  = x `div` (mypow 2 y)
intOp AbsSyn.XOR x y  = boolToVal (myxor (valToBool x) (valToBool y))
intOp op x y = error ((show op) ++" not yet implemented")

instance Evaluator Exp where
 --eval :: Env -> Expr -> (Val, Env)
 eval (CONST value) = return (I value)
 eval (NAME label) = return (L label)
 eval (TEMP temp) = do
                     x <- getRegSafe temp
                     return x
 eval c@(BINOP op left right)=do
                             lhs <- eval left
                             rhs <- eval right
                             tick (cost c)
                             return $
                               case lhs of
                                 (I x) -> case rhs of
                                           (I y) -> (I (intOp op x y))
                                           (A a) -> (A (intOp op x a))
                                 (A a) -> case rhs of
                                           (I y) -> (A (intOp op a y))
                                           (A b) -> (A (intOp op a b))
 eval c@(MEM addr) = do 
                      x <- (eval addr)
                      case x of
                       (A a) -> do
                         z <- getMemSafe a
                         tick (cost c)
                         return z
                       _ -> rtError ("Value of a MEM is not an address:\n  "++(show c)++"; val="++(show x))

 eval c@(CALL func args) = do
                            -- trace ("Entering " ++ (show c) ++ "\n") return ()
                            (L f) <- eval func
                            q <- catchExternal f args
                            case q of 
                              (Just a) -> return a
                              Nothing -> do
                            	fc <- getFunCode f
                            	ps <- getFunParams f
                            	fpReg' <- fpReg
                            	spReg' <- spReg
                            	hpReg' <- hpReg
                            	fp <- getReg fpReg'
                            	sp <- getReg spReg'
                            	hp <- getReg hpReg'
                                old_regs <- getR -- remember old register to simulate callee saving
                                -- old_mem <- getM -- NO: side effects on heap possb
                            	-- trace (".. Allocating frame; current frame at " ++ (show fp) ++"\n") return ()
                                unless ((length ps)==(length args)) $
                                  trace ("** call " ++ (show f) ++ " mismatching arg lengths: " ++ (show ps) ++ (show args)) (return ())
                            	vals <- passArgs'' ps args -- vals for tracing only
                            	rm <- getR
                            	topts <- getTraceOpts 
                            	let traceme = CallTrace `elem` topts
                            	let maps = RegMaps `elem` topts
                            	when traceme $
                            	  trace ("=> CALL " ++ (show f) ++ (show vals) ++ "\tFP= " ++ (show fp) ++ "\tSP= " ++ (show sp) ++ "\tHP= " ++ (show hp) ++ (if maps then "RegMap after passing args: " ++ (show args) ++ " -> " ++ (show ps) ++"\n" ++ (showTree rm) else "")) return ()
                                when (Debug `elem` topts) $
                                  trace ("%% BODY is\n" ++ (render (pp fc))) (return ())
                            	unless (StdCallConv `elem` topts) $
                                  allocFrame globalFrameSizeHACK
                                {-
                            	if traceme 
                            	  then getReg fpReg >>= \ fp -> getReg spReg >>= \ sp -> trace ("** new FP= " ++ (show fp) ++"\tnew SP= " ++ (show sp) ++ "\n") return ()
                                  else return ()
                                -}
                            	eval fc 
                            	rm' <- getR
                            	r <- getRegSafe retReg
                            	unless (StdCallConv `elem` topts) $
                            	  deallocFrame
                                {-
                            	if traceme 
                            	  then getReg fpReg >>= \ fp -> getReg spReg >>= \ sp -> trace ("** old FP= " ++ (show fp) ++"\told SP= " ++ (show sp) ++ "\n") return ()
                                  else return ()
                                -}
                            	when traceme $
                            	  trace ("=> CALL " ++ (show f) ++ (show vals) ++ " returns value " ++ (show r) ++ (if Debug `elem` topts then " with REG MAP\n" ++ (showTree rm') else "")) (return ())
                                -- setM old_mem
                            	unless (StdCallConv `elem` topts) $
                                  restoreCalleeSavedRegs old_regs
                                tick (cost c)
                            	return r

                         {- let
                           argvals = scanl (foldl (\ (v,q,t) x. eval t q x) (dummyVal,q,t) args
                           (q',t') = bindArgs (thd3 (last argvals)) (snd3 (last argvals)) argvals
                         in
                           eval t' q' (funTab func) -}

 eval c@(ESEQ stm exp) = tick (cost c) >> eval stm >> eval exp

 fillTab (CONST value) = return ()
 fillTab (NAME label) = return ()
 fillTab (TEMP temp) = return ()
 fillTab (BINOP op left right)= pushCtxt >> fillTab left >> popCtxt >> pushCtxt >> fillTab right >> popCtxt
 fillTab (MEM addr) =  return ()
 fillTab (CALL func args) =  pushCtxt >> fillTab args >> popCtxt
 fillTab (ESEQ stm exp) =  pushCtxt >> fillTab stm >> popCtxt >> pushCtxt >> fillTab exp >> popCtxt

 -- Cost metric
 cost (CONST value) = 0
 cost (NAME label) = 0
 cost (TEMP temp) = 0
 cost (BINOP op left right)= cost left + cost right + 1 {- depending on binop!! -}
 cost (MEM addr) = cost addr
 cost (CALL func args) = sum (Prelude.map cost args) + 3 + (toInteger (length calleeSaves)) {- depends on calling convention! -}
 cost (ESEQ stm exp) =  cost stm + cost exp 

instance Evaluator Fct where
 eval (Fct l ps ss) = error "Can't eval function declaration"
 fillTab (Fct l ps ss) = fillTab ss
 cost _ = 0

instance Evaluator [Stm] where
 eval [] =  getTraceOpts  >>= \ topts -> when (Debug `elem` topts) (trace "%% done (Stm)" (return ())) >>  return  dummyVal
 eval (x@(JUMP _ _):xs) = eval x
 eval (x@(CJUMP _ _ _ _ _):xs) = eval x
 eval (x:xs) = eval x >>  getTraceOpts  >>= \ topts -> when (Debug `elem` topts) (trace ("%% CODE now\n" ++ (render (pp x)) ++ "\nrest of CODE [Stm]\n" ++ (render (pp xs))) (return ())) >> eval xs
 fillTab [] = storeCurrBB >> initCtxt ""
 fillTab (x:xs) = fillTab x >> fillTab xs
 cost = sum . (Prelude.map cost)

instance  Evaluator [Exp] where
 eval [] =  getTraceOpts  >>= \ topts -> when (Debug `elem` topts) (trace "%% done (Stm)" (return ())) >> return dummyVal
 -- DEBUG: eval (x@(CALL _ _):xs) = trace ("%% ABOUT to CALL " ++ (render (pp x)) ++ "\nCODE now\n" ++ (render (pp x)) ++ "\nrest of CODE [Exp]\n" ++ (render (pp xs))) (return ()) >> eval x >> trace ("%% CODE now\n" ++ (render (pp x)) ++ "\nrest of CODE [Stm]\n" ++ (render (pp xs))) (return ()) >> eval xs
 eval (x:xs) = eval x >> getTraceOpts  >>= \ topts -> when (Debug `elem` topts) (trace ("%% CODE now\n" ++ (render (pp x)) ++ "\nrest of CODE [Stm]\n" ++ (render (pp xs))) (return ())) >> eval xs
 fillTab [] = storeCurrBB >> initCtxt ""
 fillTab (x:xs) = fillTab x >> fillTab xs
 cost = sum . (Prelude.map cost)
 
---------------------------------------------------------------------------
-- Mem mgmt:

data Access = InReg Register
            | InFrame (Integer, Integer) -- (Frame, Offset)
            | InMem Address
 deriving (Read,Show,Eq)

lval :: Exp -> Intp Access
lval (TEMP r) = return (InReg r)
lval (MEM e) = eval e >>= \ a -> case a of { (A a) -> return (InMem a) ; (L l) -> rtError ("lval returned label " ++ l) ; ; (I l) -> rtError ("lval returned integer " ++ (show l)) }
lval e = error ("Illegal lval in " ++ (show e))

store :: Access -> Val -> Intp Val
store (InReg r) x = getR >>= \ rm -> setR (Map.union (Map.singleton r x) rm) >>  return (dummyVal)
store (InFrame (0,o)) x = getM >>= \ mm -> fpReg >>= \ fp -> getReg fp >>= \ qq -> case qq of { (A fpa) -> setM (Map.union (Map.singleton (fpa+o) x) mm) ; (I _) -> rtError ("store: Trying to store InFrame to Int rather than Addr") } >>  return (dummyVal)
store (InFrame (1,o)) x = getM >>= \ mm -> spReg >>= \ sp -> getReg sp >>= \ qq -> case qq of { (A spa) -> setM (Map.union (Map.singleton (spa+o) x) mm) ; (I _) -> rtError ("store: Trying to store InFrame to Int rather than Addr") } >>  return (dummyVal)
store (InMem a) x = getM >>= \ mm -> fpReg >>= \ fp -> getReg fp >>= \ f -> setM (Map.union (Map.singleton a x) mm) >>  return (dummyVal)
store a x = rtError $  "store: illegal access of destination: "++(show a)

storeA (A a) x = T $ \ (e,m,lT,fT,fP,t,s,c,p) -> ((e,Map.insert a x m,lT,fT,fP,t,s,c,p), dummyVal)
storeA _ x = error "Cannot store into a non-address"

param2access :: AbsSyn.Param -> Access
param2access (LOC n) = InFrame (1,n) -- 1 means one frame off; i.e. take sp as basis in store!!
param2access (REG t) = InReg t
--param2access (ADR a) = InMem t

showMem :: Address -> Address -> Intp String
showMem from to | to <= from = return ""
showMem from to | otherwise = 
  do
    m <- getMem from
    str <- showMem (from+4{-wordSize-}) to
    return ((show from) ++ " -> " ++ (show m) ++ "\n" ++ str)

showAllMem :: Intp String
showAllMem = hpReg >>= \ hpReg' -> getReg hpReg' >>= \ x -> getSpecials >>= \ (_,_,_,_,_,hsz) ->
  let hpSt = if (hsz==0) 
               then case hpEnd of (A s) -> (s-defHpSz)
               else case hpEnd of (A s) -> (s-hsz)
  in
   case x of 
    (A xa) -> showMem hpSt xa
    (I i) -> rtError ("showAllMem: heap pointer register contains Int rather than Addr: " ++ (show (I i)))

-- NB: this is simultaneous assignment, otw old vars might be overwritten
passArgs :: [Access] -> [Exp] -> Intp [Val]
passArgs [] [] = return []
passArgs (p:ps) (e:es) = do 
	                   trace ("++ passArgs " ++ (show e) ++ " -> " ++ (show p)) (return ())
	                   showRegMap
	                   x <- eval e
                           store p x
	                   xs <- passArgs ps es
                           topts <- getTraceOpts 
                           let traceme = ParamTrace `elem` topts
                           when traceme $
                             trace ("passArgs'': " ++ (show e) ++ (show x) ++ " -> " ++ (show p)) (return ())
                           return (x:xs)

-- works only for registers now!!
passArgs' :: [Access] -> [Exp] -> Intp [Val]
-- passArgs' [] [] = return []
passArgs' ps es = do
			   vals <- mapM eval es
	                   rm <- getR
                           let alist = Prelude.zip (Prelude.map ( \ (InReg r) -> r) ps) vals
	                   setR (Map.union (Map.fromList alist) rm) -- works only for InRegs
                           topts <- getTraceOpts 
                           let traceme = ParamTrace `elem` topts
                           when traceme $
                             trace ("passArgs': " ++ (Prelude.concat (Prelude.map (\ (x,y) -> (show y) ++ " -> " ++ (show x) ++ "  ") alist))) (return ())
	                   return vals

passArgs'' :: [Access] -> [Exp] -> Intp [Val]
-- passArgs' [] [] = return []
passArgs'' ps es = do
			   vals <- mapM eval es
	                   rm <- getR
                           let alist = Prelude.zip ps vals
	                   mapM ( \ (p,v) -> store p v )  alist
                           -- store setR (Map.union (Map.fromList alist) rm) -- works only for InRegs
                           topts <- getTraceOpts 
                           let traceme = ParamTrace `elem` topts
                           when traceme $
                             trace ("passArgs': " ++ (Prelude.concat (Prelude.map (\ (x,y) -> (show y) ++ " -> " ++ (show x) ++ "  ") alist))) (return ())
	                   return vals

-- passArgs' _ _ = rtError "passArgs': Mismatch of lengths of ars and param lists"

allocFrame :: Integer ->  Intp ()
allocFrame sz = do
	         sp <- spReg
	         fp <- fpReg
              	 s@(A spa) <- getReg sp
              	 f@(A fpa) <- getReg fp
	         when (s+(I sz)<spEnd) $
                   rtError "stack overflow"
                 store (InMem spa) f -- (A fpa)
	      	 setReg fp (A spa)
	      	 setReg sp (A (spa-sz))

deallocFrame :: Intp ()
deallocFrame = do
	        sp <- spReg
	        fp <- fpReg
              	(A spa) <- getReg sp
                (A fpa) <- getReg fp
	        setReg sp (A fpa)
	        m <- getMem fpa
	        setReg fp m

initRegs :: SpecialRegs -> Intp (Map String Val)
initRegs (fp,sp,hp,tp,ssz,hsz) = 
 do 
  let spSt = if (ssz==0) 
               then spStart
               else case spEnd of (A s) -> A (s+ssz) -- -(spStart-spEnd)
  let hpSt = if (hsz==0) 
               then case hpEnd of (A s) -> A (s-defHpSz)
               else case hpEnd of (A s) -> A (s-hsz)
  let xs = Prelude.map ( \ x -> (x, dummyVal) ) calleeSaves 
  tp <- thisReg
  let rp = "t2"
  fp <- fpReg
  sp <- spReg
  hp <- hpReg
  topts <- getTraceOpts 
  when (Interp.Verbose `elem` topts) $
    trace ("Chosen Registers: TP="++tp++"; FP="++fp++"; SP="++sp++"; HP="++hp++"; RP="++rp++"; SSZ="++(show ssz)++"; HSZ="++(show hsz)++"\nSTACK:"++(show spEnd)++"-"++(show spSt)++"\nHEAP:"++(show hpSt)++"-"++(show hpEnd)) (return ())
  return $ Map.fromList ([(tp, dummyAddr),(rp, dummyInt), (fp, spSt), (sp, spSt), (hp, hpSt)]++xs)
   -- return (foldr Map.union Map.empty xs') 

-- all regs EXCEPT hp are callee saved
restoreCalleeSavedRegs :: RegMap -> Intp ()
restoreCalleeSavedRegs old_regs = 
  do 
    hpReg' <- hpReg
    hp <- getReg hpReg'
    setR old_regs
    setReg hpReg' hp
    return ()
		        

---------------------------------------------------------------------------
-- handling of external calls

catchExternal :: String -> [Exp] -> Intp (Maybe Val)
catchExternal "L_halloc_obj" es = do
                                  vals <- mapM eval es 
                                  a <- case vals of
                                    	 [(I n)] -> halloc_obj n
                                    	 _ -> rtError ("illegal arguments to _halloc_obj (should be 1 integer argument): " ++ (show vals))
                                  return (Just a)
catchExternal "L_init_obj" es = do
                                  vals <- mapM eval es 
                                  case vals of
                                    [(A a), (I n)] -> init_obj a n
                                    _ -> rtError ("illegal arguments to _init_obj (should be 1 address and 1 integer argument): " ++ (show vals))
                                  return (Just dummyAddr)
catchExternal "L_halloc_arr" es = do
                                  vals <- mapM eval es 
                                  a <- case vals of
                                    	 [(I n)] -> halloc_arr n
                                    	 _ -> rtError ("illegal arguments to _halloc_arr (should be 1 integer argument): " ++ (show vals))
                                  return (Just a)
catchExternal "L_init_arr" es = do
                                  vals <- mapM eval es 
                                  case vals of
                                    [(A a), (I n)] -> init_arr a n
                                    _ -> rtError ("illegal arguments to _init_arr (should be 1 address and 1 integer argument): " ++ (show vals))
                                  return (Just dummyAddr)
catchExternal "L_print" es = do
                                  vals <- mapM eval es 
	                          msgs (Prelude.map show vals)
                                  return (Just dummyAddr)
catchExternal _ es = return Nothing

-- allocate an object of given size
halloc_obj :: Integer -> Intp Val
halloc_obj sz = do
	          hpp <- hpReg
	          hp@(A hpa) <- getReg hpp
                  topts <- getTraceOpts 
                  let traceme = AllocTrace `elem` topts
                  when traceme $
	            trace (".. Allocating object at address " ++ (show (A hpa)) ++ "\tsize: " ++ (show sz)) (return ())
	          let new_hp = (A (hpa+sz))
	          when (new_hp>hpEnd) $
                   rtError "heap overflow"
	          hpReg' <- hpReg
	          setReg hpReg' new_hp
	          return hp

init_obj :: Address -> Integer -> Intp ()
init_obj a sz = do
	              --(A a) <- obj
	              mm <- getM
	              setM (Map.union (Map.fromList (Prelude.map (\ x -> (x,(A 0))) [a..a+sz])) mm)
	              return ()

-- allocate an array of given size
halloc_arr :: Integer -> Intp Val
halloc_arr sz = do
	          hpp <- hpReg
	          hp@(A hpa) <- getReg hpp
                  topts <- getTraceOpts 
                  let traceme = AllocTrace `elem` topts
                  when traceme $
	            trace (".. Allocating array at address " ++ (show (A hpa)) ++ "\tsize: " ++ (show sz)) (return ())
	          let new_hp = hp+(I sz)
	          when (new_hp>hpEnd) $
                    rtError "heap overflow"
	          hpReg' <- hpReg
	          setReg hpReg' new_hp
	          return (hp+(I 4))  -- NB: length at index: -1

init_arr :: Address -> Integer -> Intp ()
init_arr a sz = do
	              --(A a) <- obj
	              mm <- getM
	              setM (Map.union (Map.fromList ((a-1,(I sz)):(Prelude.map (\ x -> (x,(I 99))) [a..a+sz]))) mm)
	              return ()

-- DUMMY
addrOfFrame _ = 0

mypow m 0 = 1
mypow m n = m*(mypow m (n-1))

myxor x y = x/=y

valToBool 0 = False
valToBool _ = True

boolToVal False = 0
boolToVal True = 1
